// DataManager.cpp: implementation of the CDataManager class.
//
//////////////////////////////////////////////////////////////////////


#include "stdio.h"

#include "..\zlib\zlib.h"

#include "DataManager.h"
#include "Result.h"
#include "PhysicalFile.h"
#include "CompressFile.h"
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
namespace FileSystem
{

CResult CDataManager::initialize(char *newDataManagerPath)
{

	//initialize attributes
	ZeroMemory(dataManagerPath,sizeof(dataManagerPath));
	firstFile = NULL;
	lastFile = NULL;

	//this sould be fix laters
	if (newDataManagerPath==NULL)
		return resultOk;

	CResult result;

	//call parent method
	result = Inherit::initialize();
	
	//if was ok
	if (result==resultOk)
	{
		//if we have a data manager path
		if (newDataManagerPath!=NULL)
		{
			//store value
			strcpy(dataManagerPath,newDataManagerPath);		

			//scan for the pack files
			result = scanForPacks();
		}
	}

	//return value
	return result;
}



CResult CDataManager::addPackFile(char* fileName)
{
	CResult result = resultOk;
	
	//var to store de complete file name
	char completeName[MAX_PATH];	
	ZeroMemory(completeName,sizeof(completeName));

	//get the full name
	strcpy(completeName,dataManagerPath);
	strcat(completeName,"\\");
	strcat(completeName,fileName);

	//create the element for the list
	TPacksListElement* element = new TPacksListElement;
	
	//by default without next & prior
	element->next  = NULL;
	element->prior  = NULL;

	//if we haven't more files this its the first
	if ( lastFile == NULL)		
		firstFile = element;
	else
	{	
		//if they are any last file
		if(lastFile!=NULL )
		{
			//conect last and this element
			lastFile->next = element;			
			element->prior = lastFile;	
		}
	}

	//this element its the last one
	lastFile = element;

	//initialize the pack file element
	element->pack.initialize(completeName);
	
	//return value
	return result;
}

CResult CDataManager::scanForPacks()
{
	CResult result = resultOk;

	//var to store file search pattern
	char pattern[MAX_PATH];
	ZeroMemory(pattern,sizeof(pattern));

	//make the pattern
	strcat(pattern,dataManagerPath);
	strcat(pattern,"\\*.fpk");		


	WIN32_FIND_DATA findData;	
	HANDLE findFileHandle = NULL;

	//find first file
	findFileHandle = FindFirstFile(pattern,&findData);

	//if we get a valid search handle
	if (findFileHandle!=INVALID_HANDLE_VALUE )
	{
		//add the file
		addPackFile(findData.cFileName);

		//continue searching & adding packs
		while (FindNextFile(findFileHandle,&findData))
			addPackFile(findData.cFileName);

		//close the search
		FindClose(findFileHandle);
	}

	//return value
	return result;
}

CResult CDataManager::finalize()
{
	//quick trick
	if (strcmpi(dataManagerPath,"")==0)
		return resultOk;

	CResult result;

	result = Inherit::finalize();


	//to scroll trougth pack list
	TPacksListElement* element;
	
	//go for all of then
	while ( lastFile!=NULL)
	{
		//detach from the list
		element = lastFile;
		lastFile = lastFile->prior;

		//finalize the packfile
		element->pack.finalize();

		//delete the object
		delete element;	
	}

	return result;
}

CResult CDataManager::compressBlock(unsigned char *dest, unsigned long *destLen,
									const unsigned char *source,
									unsigned long sourceLen)
{		
	//call the zlib compres2 function & get the retun value
	int z_error = compress2(dest,destLen,source,sourceLen,9);
	
	//by the fall we get a error
	CResult result(codeFail,0,"CDataManager::compressBlock.unknow zlib error");

	switch (z_error)
	{
		//no enough memory
		case Z_MEM_ERROR:
			result = CResult(codeFail,0,
				"CDataManager::compressBlock - not enough memory");
		break;

		//no enough output buffer
		case Z_BUF_ERROR:
			result = CResult (codeFail,0,
				"CDataManager::compressBlock - not enough out buffer");
		break;

		//no error
		case Z_OK:
			result = resultOk;
		break;
	}

	//retur value
	return result;
}

CResult CDataManager::uncompressBlock(unsigned char *dest,
									  unsigned long *destLen,
									  const unsigned char *source,
									  unsigned long sourceLen)
{
	//call the zlib uncompress function & get the retun value
	int z_error = uncompress(dest,destLen,source,sourceLen);
	
	//by the fall we get a error
	CResult result(codeFail,0,
		"CDataManager::compressBlock - unknow zlib error");

	switch (z_error)
	{		
		//not enough memory
		case Z_MEM_ERROR:
			result = CResult(codeFail,0,
				"CDataManager::uncompressBlock - not enough memory");
		break;

		//not enough output buffer
		case Z_BUF_ERROR:
			result = CResult (codeFail,0,
				"CDataManager::uncompressBlock - not enough out buffer");
		break;

		//data was corrupted
		case Z_DATA_ERROR:
			result = CResult (codeFail,0,
				"CDataManager::uncompressBlock -  input data was corrupted");
		break;

		//no error
		case Z_OK:
			result = resultOk;
		 break;
	}

	return result;

}


CResult CDataManager::compressFile (char* inputFileName,char* outputFileName,
									unsigned long int blockSize,
									unsigned long int* total)
{
	CResult result = resultOk;

	//initial values
	FILE* fileInputStream=NULL;
	FILE* fileOutputStream=NULL;
	char stringBuffer[1024];


	//open file input
	fileInputStream = fopen(inputFileName,"rb");

	//if we couldn't open it
	if (fileInputStream==NULL)
	{
		//error string
		sprintf(stringBuffer,
			"CDataManager::compressFile - couldn't open file %s",inputFileName);

		result = CResult(codeFail,0,stringBuffer);

	}
	else
	{
		//open out stream
		fileOutputStream = fopen(outputFileName,"wb");

		//if wee couldn't open it
		if (fileInputStream==NULL)
		{
			//error string
			sprintf(stringBuffer,
				"CDataManager::compressFile - couldn't open file %s",
				outputFileName);

			result = CResult(codeFail,0,stringBuffer);	
		}
		else			
			//compres the file
			result = compressFile(fileInputStream,
						fileOutputStream,blockSize,total);
	}

	//close the files
	if (fileInputStream!=NULL)
		fclose(fileInputStream);	
	if (fileOutputStream!=NULL)
		fclose(fileOutputStream);

	//return value
	return result;
}

CResult CDataManager::uncompressFile(char *inputFileName, char *outputFileName,
									 unsigned long blockSize,
									 unsigned long int total)
{	
	CResult result = resultOk;

    //initial values
	FILE* fileInputStream=NULL;
	FILE* fileOutputStream=NULL;
	char stringBuffer[1024];

	//open input stream
	fileInputStream = fopen(inputFileName,"rb");

	//if we couldn't open it
	if (fileInputStream==NULL)
	{	
		//error string
		sprintf(stringBuffer,
			"CDataManager::uncompressFile - couldn't open file %s",
			inputFileName);

		result = CResult(codeFail,0,stringBuffer);
	}
	else
	{
		//open output stream
		fileOutputStream = fopen(outputFileName,"wb");

		//if we couldn't open it
		if (fileInputStream==NULL)
		{
			//error string
			sprintf(stringBuffer,
				"CDataManager::uncompressFile - couldn't open file %s",
				outputFileName);

			result = CResult(codeFail,0,stringBuffer);		
		}
		else
			//uncompres the file
			result = uncompressFile (fileInputStream,fileOutputStream,
			blockSize,total);	
	}

	//close the files
	if (fileInputStream!=NULL) 
		fclose(fileInputStream);
	if (fileOutputStream!=NULL) 
		fclose(fileOutputStream);

	//result value
	return result;

}


CResult CDataManager::uncompressFile (FILE* fileInputStream,
									  FILE* fileOutputStream,
									  unsigned long int blockSize,
									  unsigned long int total)
{

	CResult result = resultOk;

	//init vars
	unsigned long int subtotal=0;
	BYTE* inputBlock = NULL;		
	BYTE* outputBlock= NULL;
	unsigned long int readedSize = 0;
	unsigned long int compressSize = 0;

	//data buffers
	inputBlock = new BYTE[blockSize+12];
	outputBlock = new BYTE[blockSize];

	//until eof or amount read, or error happens
	while(!feof(fileInputStream)  && (result == resultOk) && (subtotal<total) )
	{
		//zeros to the buffers
		ZeroMemory(inputBlock,blockSize+12);
		ZeroMemory(outputBlock,blockSize);

		//read the compres/uncompres size of the data
		fread(&readedSize,1,sizeof(readedSize),fileInputStream);
		fread(&compressSize,1,sizeof(compressSize),fileInputStream);

		//increase amount read from the input buffer
		subtotal += sizeof(compressSize);
		subtotal += sizeof(readedSize);
		subtotal += compressSize;			

		//read the amount of data that its compresed
		fread(inputBlock,1,compressSize,fileInputStream);

		//uncompress the data into the out buffer
		result = uncompressBlock(outputBlock,&readedSize,inputBlock,
			compressSize);

		//if uncompreses was correct
		if (result==resultOk)
		{
			//write the uncompress data
			fwrite(outputBlock,1,readedSize,fileOutputStream);
		}	
	}

	//free the bufers
	delete [] inputBlock;
	delete [] outputBlock;

	//return result
	return result;
}

CResult CDataManager::compressFile (FILE* fileInputStream,
									FILE* fileOutputStream,
									unsigned long int blockSize,
									unsigned long int* total)
{
	CResult result = resultOk;

	//by default 0 output bytes
	*total = 0;

	//init vars
	BYTE* inputBlock = NULL;		
	BYTE* outputBlock= NULL;
	unsigned long int readedSize = 0;
	unsigned long int compressSize = 0;

	//data buffers
	inputBlock = new BYTE[blockSize];
	outputBlock = new BYTE[blockSize+12];

	//until eof or an error happens
	while(!feof(fileInputStream) && (result == resultOk) )
	{
		//set zeroes to the buffers
		ZeroMemory(inputBlock,blockSize);
		ZeroMemory(outputBlock,blockSize+12);

		//read input buffer
		readedSize = fread(inputBlock,1,blockSize,fileInputStream);

		//increase the posible compress size
		compressSize = readedSize +12;

		//compress the block
		result = compressBlock (outputBlock,&compressSize,inputBlock,
						readedSize);

		//write the uncompress size
		fwrite(&readedSize,1,sizeof(readedSize),fileOutputStream);

		//write the compress size
		fwrite(&compressSize,1,sizeof(compressSize),fileOutputStream);

		//write the buffer od compress data
		fwrite(outputBlock,1,compressSize,fileOutputStream);
	
		//increase total
		*total += sizeof(compressSize);
		*total += sizeof(readedSize);
		*total += compressSize;
	}
	
	//free the buffers
	delete [] inputBlock;
	delete [] outputBlock;

	//return values
	return result;
}


CDataFile* CDataManager::openFile(const char *filePath)
{
	//by default returns NULL 
	CDataFile* file = NULL;

	//try physicall file
	char strBuff[260];
	ZeroMemory(strBuff,sizeof(strBuff));

	//full path file name
	strcpy( strBuff, dataManagerPath);
	strcat( strBuff, filePath);

	//could we opened?
	FILE* fileHandle = fopen(strBuff,"rb");

	//open ok
	if ( fileHandle != NULL)
	{
		//close the file
		fclose (fileHandle);

		//set the physical file
		file = new CPhysicalFile();
		((CPhysicalFile*)file)->initialize(this,strBuff);
		
	}
	else
	{
		TPacksListElement* packElement = firstFile;
		TFileInfoElement* element = NULL;

		while ((packElement!=NULL)&&(file==NULL))
		{
			element = packElement->pack.getFirstFile();
			while(element!=NULL)
			{
				if (strcmpi(filePath,element->info.fileName)==0)						
				{
					//set the compress file
					file = new CCompressFile();
					((CCompressFile*)file)->initialize(this,packElement,element);
				}

				element = element->next;
			}

			packElement = packElement->next;
		}
	}

	return file;
}

CResult CDataManager::closeFile(CDataFile *filePointer)
{
	CResult result = resultOk;

	//if we have a NULL pointer skip
	if ( filePointer!= NULL)
	{
		//finalize the file and delete the object
		result = filePointer->finalize();
		delete filePointer;
	}
	else
	{
		//report string
		result = CResult(codeFail,0,"Error try to close a null file");
	}

	//return value
	return result;
}

}

//for the No object oriented FileSystem
FileSystem::CDataManager FS;

bool InitFS(char* path)
{	  	
	//just init the FS object
	return (FS.initialize(path)==FileSystem::resultOk);;
}

FSFILE* fsopen(const char* name)
{
  //open whith the FS and return
  return FS.openFile(name);
}


void fsclose(FSFILE* file)
{
  //close with FS
  FS.closeFile(file);
}

int fsseek(FSFILE* file,long amount,int pos)
{
  //just seek the object
  return file->seek(amount,pos);
}

int fsread(long amount,char* buffer,FSFILE* file)
{
  //just read from object
  return file->read(amount,buffer);
}

char* fsgets(char* buffer,int max,FSFILE* file)
{
  //just get string from object
  return file->gets(buffer,max);
}

int fseof(FSFILE* file)
{
  //just get eof object
  return file->eof();
}


